from cqi import CQiClient
from cqi.errors import CQiException
from cqi.status import CQiStatus
from flask import session
from inspect import signature
from threading import Lock
from typing import Callable, Dict, List
from app import socketio
from app.decorators import socketio_login_required
from . import NAMESPACE as ns
from .extensions import (
    corpus_update_db,
    corpus_static_data,
    corpus_paginate_corpus,
    cqp_paginate_subcorpus,
    cqp_partial_export_subcorpus,
    cqp_export_subcorpus,
)


CQI_API_FUNCTIONS: List[str] = [
    'ask_feature_cl_2_3',
    'ask_feature_cqi_1_0',
    'ask_feature_cqp_2_3',
    'cl_alg2cpos',
    'cl_attribute_size',
    'cl_cpos2alg',
    'cl_cpos2id',
    'cl_cpos2lbound',
    'cl_cpos2rbound',
    'cl_cpos2str',
    'cl_cpos2struc',
    'cl_drop_attribute',
    'cl_id2cpos',
    'cl_id2freq',
    'cl_id2str',
    'cl_idlist2cpos',
    'cl_lexicon_size',
    'cl_regex2id',
    'cl_str2id',
    'cl_struc2cpos',
    'cl_struc2str',
    'corpus_alignment_attributes',
    'corpus_charset',
    'corpus_drop_corpus',
    'corpus_full_name',
    'corpus_info',
    'corpus_list_corpora',
    'corpus_positional_attributes',
    'corpus_properties',
    'corpus_structural_attribute_has_values',
    'corpus_structural_attributes',
    'cqp_drop_subcorpus',
    'cqp_dump_subcorpus',
    'cqp_fdist_1',
    'cqp_fdist_2',
    'cqp_list_subcorpora',
    'cqp_query',
    'cqp_subcorpus_has_field',
    'cqp_subcorpus_size',
    'ctrl_bye',
    'ctrl_connect',
    'ctrl_last_general_error',
    'ctrl_ping',
    'ctrl_user_abort'
]
CQI_NOPAQUE_FUNCTIONS: Dict[str, Callable] = {
    'nopaque_corpus_update_db': corpus_update_db,
    'nopaque_corpus_static_data': corpus_static_data,
    'nopaque_corpus_paginate_corpus': corpus_paginate_corpus,
    'nopaque_cqp_paginate_subcorpus': cqp_paginate_subcorpus,
    'nopaque_cqp_partial_export_subcorpus': cqp_partial_export_subcorpus,
    'nopaque_cqp_export_subcorpus': cqp_export_subcorpus,
}


@socketio.on('cqi', namespace=ns)
@socketio_login_required
def cqi_over_sio(fn_data):
    try:
        fn_name: str = fn_data['fn_name']
    except KeyError:
        return {'code': 400, 'msg': 'Bad Request'}
    fn_name: str = fn_data['fn_name']
    fn_args: Dict = fn_data.get('fn_args', {})
    try:
        cqi_client: CQiClient = session['cqi_over_sio']['cqi_client']
        cqi_client_lock: Lock = session['cqi_over_sio']['cqi_client_lock']
    except KeyError:
        return {'code': 424, 'msg': 'Failed Dependency'}
    if fn_name in CQI_API_FUNCTIONS:
        fn: Callable = getattr(cqi_client.api, fn_name)
    elif fn_name in CQI_NOPAQUE_FUNCTIONS:
        fn_args['cqi_client'] = cqi_client
        fn: Callable = CQI_NOPAQUE_FUNCTIONS[fn_name]
    else:
        return {'code': 400, 'msg': 'Bad Request'}
    for param in signature(fn).parameters.values():
        if param.default is param.empty:
            if param.name not in fn_args:
                return {'code': 400, 'msg': 'Bad Request'}
        else:
            if param.name not in fn_args:
                continue
        if type(fn_args[param.name]) is not param.annotation:
            return {'code': 400, 'msg': 'Bad Request'}
    cqi_client_lock.acquire()
    try:
        return_value = fn(**fn_args)
    except BrokenPipeError:
        return_value = {
            'code': 500,
            'msg': 'Internal Server Error'
        }
    except CQiException as e:
        return_value = {
            'code': 502,
            'msg': 'Bad Gateway',
            'payload': {
                'code': e.code,
                'desc': e.description,
                'msg': e.__class__.__name__
            }
        }
    finally:
        cqi_client_lock.release()
    if isinstance(return_value, CQiStatus):
        payload = {
            'code': return_value.code,
            'msg': return_value.__class__.__name__
        }
    else:
        payload = return_value
    return {'code': 200, 'msg': 'OK', 'payload': payload}
