from collections import Counter
from cqi import CQiClient
from cqi.status import StatusOk
from flask import session
from typing import Dict
import json
import math
import os
from app import db
from app.models import Corpus
from .utils import lookups_by_cpos, export_subcorpus, partial_export_subcorpus


def corpus_update_db(cqi_client: CQiClient, corpus: str):
    db_corpus = Corpus.query.get(session['cqi_over_sio']['corpus_id'])
    cqi_corpus = cqi_client.corpora.get(corpus)
    db_corpus.num_tokens = cqi_corpus.size
    db.session.commit()
    return StatusOk()


def corpus_static_data(cqi_client: CQiClient, corpus: str) -> Dict:
    db_corpus = Corpus.query.get(session['cqi_over_sio']['corpus_id'])
    static_corpus_data_file = os.path.join(db_corpus.path, 'cwb', 'static.json')
    if os.path.exists(static_corpus_data_file):
        with open(static_corpus_data_file, 'r') as f:
            return json.load(f)
    cqi_corpus = cqi_client.corpora.get(corpus)
    ##########################################################################
    # A faster way to get cpos boundaries for smaller s_attrs                #
    ##########################################################################
    # cqi_corpus.query('Last', '<s> []* </s>;')
    # cqi_subcorpus = cqi_corpus.subcorpora.get('Last')
    # print(cqi_subcorpus.size)
    # first_match = 0
    # last_match = cqi_subcorpus.attrs['size'] - 1
    # match_boundaries = zip(
    #     list(range(first_match, last_match + 1)),
    #     cqi_subcorpus.dump(cqi_subcorpus.attrs['fields']['match'], first_match, last_match),
    #     cqi_subcorpus.dump(cqi_subcorpus.attrs['fields']['matchend'], first_match, last_match)
    # )
    # for x in match_boundaries:
    #     print(x)
    cqi_p_attrs = {
        p_attr.name: p_attr
        for p_attr in cqi_corpus.positional_attributes.list()
    }
    cqi_s_attrs = {
        s_attr.name: s_attr
        for s_attr in cqi_corpus.structural_attributes.list()
    }
    static_corpus_data = {
        'corpus': {
            'bounds': [0, cqi_corpus.size - 1],
            'counts': {
                'token': cqi_corpus.size
            },
            'freqs': {}
        },
        'p_attrs': {},
        's_attrs': {},
        'values': {'p_attrs': {}, 's_attrs': {}}
    }
    for p_attr in cqi_p_attrs.values():
        static_corpus_data['corpus']['freqs'][p_attr.name] = dict(
            zip(
                range(0, p_attr.lexicon_size),
                p_attr.freqs_by_ids(list(range(0, p_attr.lexicon_size)))
            )
        )
        static_corpus_data['p_attrs'][p_attr.name] = dict(
            zip(
                range(0, cqi_corpus.size),
                p_attr.ids_by_cpos(list(range(0, cqi_corpus.size)))
            )
        )
        static_corpus_data['values']['p_attrs'][p_attr.name] = dict(
            zip(
                range(0, p_attr.lexicon_size),
                p_attr.values_by_ids(list(range(0, p_attr.lexicon_size)))
            )
        )
    for s_attr in cqi_s_attrs.values():
        if s_attr.has_values:
            continue
        static_corpus_data['corpus']['counts'][s_attr.name] = s_attr.size
        static_corpus_data['s_attrs'][s_attr.name] = {'lexicon': {}, 'values': None}
        static_corpus_data['values']['s_attrs'][s_attr.name] = {}
        for id in range(0, s_attr.size):
            static_corpus_data['s_attrs'][s_attr.name]['lexicon'][id] = {}
            lbound, rbound = s_attr.cpos_by_id(id)
            static_corpus_data['s_attrs'][s_attr.name]['lexicon'][id]['bounds'] = [lbound, rbound]
            static_corpus_data['s_attrs'][s_attr.name]['lexicon'][id]['counts'] = {}
            static_corpus_data['s_attrs'][s_attr.name]['lexicon'][id]['counts']['token'] = rbound - lbound + 1
            if s_attr.name not in ['text', 's']:
                continue
            cpos_range = range(lbound, rbound + 1)
            static_corpus_data['s_attrs'][s_attr.name]['lexicon'][id]['counts']['ent'] = len({x for x in cqi_s_attrs['ent'].ids_by_cpos(list(cpos_range)) if x != -1})
            if s_attr.name != 'text':
                continue
            static_corpus_data['s_attrs'][s_attr.name]['lexicon'][id]['counts']['s'] = len({x for x in cqi_s_attrs['s'].ids_by_cpos(list(cpos_range)) if x != -1})
            static_corpus_data['s_attrs'][s_attr.name]['lexicon'][id]['freqs'] = {}
            for p_attr in cqi_p_attrs.values():
                static_corpus_data['s_attrs'][s_attr.name]['lexicon'][id]['freqs'][p_attr.name] = dict(Counter(p_attr.ids_by_cpos(list(cpos_range))))
        sub_s_attrs = cqi_corpus.structural_attributes.list(filters={'part_of': s_attr})
        s_attr_value_names = [
            sub_s_attr.name[(len(s_attr.name) + 1):]
            for sub_s_attr in sub_s_attrs
        ]
        sub_s_attr_values = [
            sub_s_attr.values_by_ids(list(range(0, s_attr.size)))
            for sub_s_attr in sub_s_attrs
        ]
        static_corpus_data['s_attrs'][s_attr.name]['values'] = s_attr_value_names
        static_corpus_data['values']['s_attrs'][s_attr.name] = {
            s_attr_id: {
                s_attr_value_name: sub_s_attr_values[s_attr_value_name_idx][s_attr_id_idx]
                for s_attr_value_name_idx, s_attr_value_name in enumerate(
                    static_corpus_data['s_attrs'][s_attr.name]['values']
                )
            } for s_attr_id_idx, s_attr_id in enumerate(range(0, s_attr.size))
        }
    with open(static_corpus_data_file, 'w') as f:
        json.dump(static_corpus_data, f)
    return static_corpus_data


def corpus_paginate_corpus(
    cqi_client: CQiClient,
    corpus: str,
    page: int = 1,
    per_page: int = 20
) -> Dict:
    cqi_corpus = cqi_client.corpora.get(corpus)
    # Sanity checks
    if (
        per_page < 1
        or page < 1
        or (
            cqi_corpus.size > 0
            and page > math.ceil(cqi_corpus.size / per_page)
        )
    ):
        return {'code': 416, 'msg': 'Range Not Satisfiable'}
    first_cpos = (page - 1) * per_page
    last_cpos = min(cqi_corpus.size, first_cpos + per_page)
    cpos_list = [*range(first_cpos, last_cpos)]
    lookups = lookups_by_cpos(cqi_corpus, cpos_list)
    payload = {}
    # the items for the current page
    payload['items'] = [cpos_list]
    # the lookups for the items
    payload['lookups'] = lookups
    # the total number of items matching the query
    payload['total'] = cqi_corpus.size
    # the number of items to be displayed on a page.
    payload['per_page'] = per_page
    # The total number of pages
    payload['pages'] = math.ceil(payload['total'] / payload['per_page'])
    # the current page number (1 indexed)
    payload['page'] = page if payload['pages'] > 0 else None
    # True if a previous page exists
    payload['has_prev'] = payload['page'] > 1 if payload['page'] else False
    # True if a next page exists.
    payload['has_next'] = payload['page'] < payload['pages'] if payload['page'] else False  # noqa
    # Number of the previous page.
    payload['prev_num'] = payload['page'] - 1 if payload['has_prev'] else None
    # Number of the next page
    payload['next_num'] = payload['page'] + 1 if payload['has_next'] else None
    return payload


def cqp_paginate_subcorpus(
    cqi_client: CQiClient,
    subcorpus: str,
    context: int = 50,
    page: int = 1,
    per_page: int = 20
) -> Dict:
    corpus_name, subcorpus_name = subcorpus.split(':', 1)
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_subcorpus = cqi_corpus.subcorpora.get(subcorpus_name)
    # Sanity checks
    if (
        per_page < 1
        or page < 1
        or (
            cqi_subcorpus.size > 0
            and page > math.ceil(cqi_subcorpus.size / per_page)
        )
    ):
        return {'code': 416, 'msg': 'Range Not Satisfiable'}
    offset = (page - 1) * per_page
    cutoff = per_page
    cqi_results_export = export_subcorpus(
        cqi_subcorpus, context=context, cutoff=cutoff, offset=offset)
    payload = {}
    # the items for the current page
    payload['items'] = cqi_results_export.pop('matches')
    # the lookups for the items
    payload['lookups'] = cqi_results_export
    # the total number of items matching the query
    payload['total'] = cqi_subcorpus.size
    # the number of items to be displayed on a page.
    payload['per_page'] = per_page
    # The total number of pages
    payload['pages'] = math.ceil(payload['total'] / payload['per_page'])
    # the current page number (1 indexed)
    payload['page'] = page if payload['pages'] > 0 else None
    # True if a previous page exists
    payload['has_prev'] = payload['page'] > 1 if payload['page'] else False
    # True if a next page exists.
    payload['has_next'] = payload['page'] < payload['pages'] if payload['page'] else False  # noqa
    # Number of the previous page.
    payload['prev_num'] = payload['page'] - 1 if payload['has_prev'] else None
    # Number of the next page
    payload['next_num'] = payload['page'] + 1 if payload['has_next'] else None
    return payload


def cqp_partial_export_subcorpus(
    cqi_client: CQiClient,
    subcorpus: str,
    match_id_list: list,
    context: int = 50
) -> Dict:
    corpus_name, subcorpus_name = subcorpus.split(':', 1)
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_subcorpus = cqi_corpus.subcorpora.get(subcorpus_name)
    cqi_subcorpus_partial_export = partial_export_subcorpus(cqi_subcorpus, match_id_list, context=context)
    return cqi_subcorpus_partial_export


def cqp_export_subcorpus(
    cqi_client: CQiClient,
    subcorpus: str,
    context: int = 50
) -> Dict:
    corpus_name, subcorpus_name = subcorpus.split(':', 1)
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_subcorpus = cqi_corpus.subcorpora.get(subcorpus_name)
    cqi_subcorpus_export = export_subcorpus(cqi_subcorpus, context=context)
    return cqi_subcorpus_export
